/*
 * $Id: BNF.h,v 1.2 2009/12/18 13:55:18 jpt Exp $
 * Date (yymmdd):               090127
 * Version:                     0.1
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */

#ifndef _BNF_H
#define	_BNF_H

#include <math.h>
#include <string>
#include <cassert>
#include <sstream>
#include <fstream>
#include <iostream>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <errno.h>
#include <boost/spirit/include/classic_core.hpp>
#include <boost/spirit/include/classic_ast.hpp>
#include <libxml++/libxml++.h>

using namespace boost::spirit::classic;
using namespace xmlpp;
using namespace std;

namespace ECSB {

// Here's some typedefs to simplify things
typedef char const*                              iterator_t;
typedef tree_match<iterator_t>                   parse_tree_match_t;
typedef parse_tree_match_t::const_tree_iterator  iter_t;

////////////////////////////////////////////////////////////////////////////
//  Our P system BNF grammar
////////////////////////////////////////////////////////////////////////////
struct PSModel: public grammar<PSModel>
{
    template <typename ScannerT>
    struct definition
    {
        definition(PSModel const& /*self*/)
        {

            Psystem = discard_node_d[str_p("SPsystem")] >> root_node_d[PsystemName] >>
                        alphabet >>
                        compartments >>
                        initialMultisets >>
                        ruleSets >>
                      discard_node_d[str_p("endSPsystem")];

            PsystemName = leaf_node_d[ lexeme_d[(+alnum_p)] ];

            alphabet = root_node_d[str_p("alphabet")] >>
                            +(objectName) >>
                       discard_node_d[ str_p("endAlphabet") ];

            objectName = leaf_node_d[ (lexeme_d[(+(alnum_p | ch_p('_')))] - (str_p("endAlphabet") | str_p("endInitialMultiset")))];

            compartments = root_node_d[str_p("compartments")] >>
                                +(compartment) >>
                           discard_node_d[str_p("endCompartments")];

            //compartment = elementary | nonElementary;

            //elementary = root_node_d[compartmentName];

            compartment = root_node_d[compartmentName] >> !(discard_node_d[ str_p("inside") ] >> compartmentName);

            compartmentName = leaf_node_d[ (lexeme_d[ ((+alnum_p) - (str_p("in") | str_p("endCompartments"))) ])]; //leaf_node_d[ lexeme_d[ +alnum_p]];

            initialMultisets = root_node_d[str_p("initialMultisets")] >>
                                    +(initialMultiset) >>
                               discard_node_d[str_p("endInitialMultisets")];

            initialMultiset = discard_node_d[str_p("initialMultiset")] >> root_node_d[compartmentName] >>
                                    *(multiplicity) >>
                              discard_node_d[str_p("endInitialMultiset")];

            multiplicity = root_node_d[objectName] >> int_p;

            //int_p = leaf_node_d[ lexeme_d[+digit_p] ];

            ruleSets = root_node_d[str_p("ruleSets")] >>
                            *(ruleSet) >>
                       discard_node_d[str_p("endRuleSets")];

            ruleSet = discard_node_d[str_p("ruleSet")] >> root_node_d[compartmentName] >>
                        *(PSRule | module) >>
                      discard_node_d[str_p("endRuleSet")];

            PSRule = BoundaryRule | TranslRule;

            BoundaryRule = root_node_d[ruleName] >> ch_p(':') >>
                           LHS  >>
                           ch_p('-') >> constantName >> leaf_node_d[ lexeme_d[str_p("->")]] >>
                           RHS >>
                           ruleConstant;

            ruleName = leaf_node_d[ (lexeme_d[(+alnum_p)] - (str_p("endRuleSet") | str_p("module"))) ];

            LHS =  LHS_outside >>  LHS_inside >> labelName;
            LHS_outside = multiset >> root_node_d[ch_p('[')];
            LHS_inside = multiset >> root_node_d[str_p("]_")];

            RHS = RHS_outside >> RHS_inside >> labelName;
            RHS_outside = multiset >> root_node_d[ch_p('[')];
            RHS_inside = multiset >> root_node_d[str_p("]_")];

            ruleConstant = constantName >> root_node_d[ch_p('=')] >> real_p;

            multiset = !(objectName >> *(root_node_d[ch_p('+')] >> objectName));

            labelName =  leaf_node_d[ lexeme_d[(+alnum_p)] ];

            constantName = leaf_node_d[ lexeme_d[+(alnum_p | ch_p('.') | ch_p('_'))] ];

            TranslRule = root_node_d[ruleName] >> ch_p(':') >>
                         LHSTransl  >> ch_p('[') >> ch_p(']') >>
                         ch_p('-') >> constantName >> str_p("->") >>
                         RHSTransl >>
                         ruleConstant;

            LHSTransl = LHS_objects >> labelName >> ruleVector;

            LHS_objects = root_node_d[ch_p('[')] >> multiset >> discard_node_d[ str_p("]_")];
            ruleVector = discard_node_d[str_p("=")] >> discard_node_d[str_p("(")] >>
                    int_p >> root_node_d[str_p(",")] >> int_p >>
                    discard_node_d[str_p(")")] >> discard_node_d[str_p("=")];

            RHSTransl =  discard_node_d[str_p("[")] >> discard_node_d[str_p("]_")] >>
                         labelName >>  ruleVector >> RHS_objects ;
            RHS_objects = root_node_d[ch_p('[')] >> multiset >> discard_node_d[ str_p("]")];


            module = root_node_d[moduleName] >> discard_node_d[ str_p("(") ] >>
                        moduleObjects >>
                        discard_node_d[ str_p(",") ] >>
                        moduleConstants >>
                        discard_node_d[ str_p(",") ] >>
                        moduleCompartments >>
                    discard_node_d[ str_p(")") ] >>
                    discard_node_d[str_p("from")] >> libraryName;

/*
            module = discard_node_d[ str_p("module") ] >> root_node_d[moduleName] >> discard_node_d[str_p("from")] >> libraryName >>
                        moduleObjects >>
                        moduleConstants >>
                        moduleCompartments >>
                     discard_node_d[ str_p("endModule") ];
*/

            moduleName = leaf_node_d[ lexeme_d[(+alnum_p)] ];

            libraryName = leaf_node_d[ lexeme_d[+(alnum_p | ch_p('.'))] ];

            moduleObjects = root_node_d[ str_p("{") ] >>
                                !(objectName >> *((discard_node_d[ch_p(',')] >> objectName))) >>
                            discard_node_d[ str_p("}") ];

            moduleConstants = root_node_d[ str_p("{") ] >>
                                !(constantName >> *(discard_node_d[ch_p(',')] >> constantName)) >>
                              discard_node_d[ str_p("}") ];

            moduleCompartments = root_node_d[ str_p("{") ] >>
                                    !(labelName >> *((discard_node_d[ch_p(',')] >> labelName))) >>
                                 discard_node_d[ str_p("}") ];

/*
            moduleObjects = root_node_d[ str_p("objects") ] >> discard_node_d[ ch_p('=') ] >>
                            !(objectName >> *((discard_node_d[ch_p(',')] >> objectName)));

            moduleConstants = root_node_d[ str_p("constants") ] >> discard_node_d[ ch_p('=') ] >>
                              !(constantName >> *(discard_node_d[ch_p(',')] >> constantName));

            moduleCompartments = root_node_d[str_p("compartments")] >> discard_node_d[ch_p('=')] >>
                                !(labelName >> *((discard_node_d[ch_p(',')] >> labelName)));
*/

        }

        rule<ScannerT> Psystem, PsystemName, alphabet, objectName, compartments, compartment, elementary, nonElementary, initialMultisets, initialMultiset, multiplicity, ruleSets, ruleSet, compartmentName, PSRule, BoundaryRule, LHS, LHS_outside, LHS_inside, RHS, RHS_outside, RHS_inside, ruleConstant, TranslRule, LHSTransl, LHS_objects, ruleVector, RHSTransl, RHS_objects, multiset, labelName, constantName, ruleName, module, moduleName, libraryName, moduleObjects, moduleConstants, moduleCompartments;

        rule<ScannerT> const&
        start() const { return Psystem; }
    };
};


////////////////////////////////////////////////////////////////////////////
//  Our Library BNF grammar
////////////////////////////////////////////////////////////////////////////
struct library_p: public grammar<library_p>
{
    template <typename ScannerT>
    struct definition
    {
        definition(library_p const& /*self*/)
        {

            Library = discard_node_d[str_p("libraryOfModules")] >> root_node_d[libraryName] >>
                        +moduleDefinition >>
                      discard_node_d[str_p("endLibraryOfModules")];

            libraryName = leaf_node_d[lexeme_d[+(alnum_p | ch_p('_'))]];//     leaf_node_d[ lexeme_d[+(alnum_p | ch_p('_'))] ];

            moduleDefinition =  root_node_d[moduleName] >> discard_node_d[ str_p("(") ] >>
                                    moduleObjects >>
                                    discard_node_d[ str_p(",") ] >>
                                    moduleConstants >>
                                    discard_node_d[ str_p(",") ] >>
                                    moduleCompartments >>
                                discard_node_d[ str_p(")") ] >>
                    discard_node_d[ str_p("=") ] >>
                    ruleSet;

            moduleName = leaf_node_d[ lexeme_d[(+alnum_p)] ];

            moduleObjects = root_node_d[ str_p("{") ] >>
                                !(objectName >> *((discard_node_d[ch_p(',')] >> objectName))) >>
                            discard_node_d[ str_p("}") ];

            moduleConstants = root_node_d[ str_p("{") ] >>
                                !(constantName >> *(discard_node_d[ch_p(',')] >> constantName)) >>
                              discard_node_d[ str_p("}") ];

            moduleCompartments = root_node_d[ str_p("{") ] >>
                                    !(labelName >> *((discard_node_d[ch_p(',')] >> labelName))) >>
                                 discard_node_d[ str_p("}") ];

            moduleConstant = constantName | constantBounds;

            constantBounds = root_node_d[constantName] >> 
                             real_p >> discard_node_d[str_p(":")] >> real_p >> discard_node_d[str_p(":")] >> real_p >>
                             scale;
            scale = str_p("linear") | str_p("logarithmic");

            ruleSet = root_node_d[str_p("{")] >> 
                        *(PSRule | module) >>
                      discard_node_d[str_p("}")];

            PSRule = BoundaryRule | TranslRule;

            BoundaryRule = root_node_d[ruleName] >> ':' >>
                     LHS  >>
                     ch_p('-') >> constantName >> leaf_node_d[ lexeme_d[str_p("->")]] >>
                     RHS >>
                     !ruleConstant;

            ruleName = leaf_node_d[ (lexeme_d[(+alnum_p)] - (str_p("endRuleSet") | str_p("module"))) ];

            LHS =  LHS_outside >>  LHS_inside >> labelName;
            LHS_outside = multiset >> root_node_d[ch_p('[')];
            LHS_inside = multiset >> root_node_d[str_p("]_")];

            RHS = RHS_outside >> RHS_inside >> labelName;
            RHS_outside = multiset >> root_node_d[ch_p('[')];
            RHS_inside = multiset >> root_node_d[str_p("]_")];

            multiset = !(objectName >> *(root_node_d[ch_p('+')] >> objectName));

            labelName =  leaf_node_d[ lexeme_d[(+alnum_p)] ];

            constantName = leaf_node_d[ lexeme_d[+(alnum_p | ch_p('.') | ch_p('_'))]];

            ruleConstant = constantName >> root_node_d[str_p("=")] >> real_p;

            objectName = leaf_node_d[ (lexeme_d[(+(alnum_p | ch_p('_')))])];

            TranslRule = root_node_d[ruleName] >> ch_p(':') >>
                         LHSTransl  >> ch_p('[') >> ch_p(']') >>
                         ch_p('-') >> constantName >> str_p("->") >>
                         RHSTransl >>
                         !ruleConstant;

            LHSTransl = LHS_objects >> labelName >> ruleVector;

            LHS_objects = root_node_d[ch_p('[')] >> multiset >> discard_node_d[ str_p("]_")];
            ruleVector = discard_node_d[str_p("=")] >> discard_node_d[str_p("(")] >>
                    int_p >> root_node_d[str_p(",")] >> int_p >>
                    discard_node_d[str_p(")")] >> discard_node_d[str_p("=")];

            RHSTransl =  discard_node_d[str_p("[")] >> discard_node_d[str_p("]_")] >>
                         labelName >>  ruleVector >> RHS_objects ;
            RHS_objects = root_node_d[ch_p('[')] >> multiset >> discard_node_d[ str_p("]")];

            module = root_node_d[moduleName] >> discard_node_d[ str_p("(") ] >>
                        moduleObjects >>
                        discard_node_d[ str_p(",") ] >>
                        moduleConstants >>
                        discard_node_d[ str_p(",") ] >>
                        moduleCompartments >>
                    discard_node_d[ str_p(")") ] >>
                    discard_node_d[str_p("from")] >> libraryName;

            libraryName = leaf_node_d[ lexeme_d[+(alnum_p | ch_p('.'))] ];

        }

        rule<ScannerT> Library, libraryName, moduleDefinition, moduleName, moduleObjects, moduleConstants, moduleCompartments, moduleConstant, constantBounds, scale, ruleSet, PSRule, BoundaryRule, ruleName, LHS, LHS_outside, LHS_inside, RHS, RHS_outside, RHS_inside, labelName, multiset, constantName, ruleConstant, objectName, TranslRule, LHSTransl, LHS_objects, ruleVector, RHSTransl, RHS_objects, module;

        rule<ScannerT> const&
        start() const { return Library; }
    };
};


////////////////////////////////////////////////////////////////////////////
//  Our Library BNF grammar
////////////////////////////////////////////////////////////////////////////
struct lattice_p: public grammar<lattice_p>
{
    template <typename ScannerT>
    struct definition
    {
        definition(lattice_p const&)
        {

            Lattice = discard_node_d[str_p("lattice")] >> root_node_d[latticeName] >>
                        dimension >>
                        !(scale) >>
                        !(units) >>
                        xmin >>
                        xmax >>
                        ymin >>
                        ymax >>
                        !(parameters) >>
                        basis >>
                        !(vertices) >>
                        neighbours >>
                      discard_node_d[str_p("endLattice")];

            latticeName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            dimension = root_node_d[str_p("dimension")] >> ( ch_p('1') | ch_p('2') );

            scale = root_node_d[str_p("scale")] >> int_p;

            units = root_node_d[str_p("units")] >> unitName;

            unitName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            xmin = root_node_d[str_p("xmin")] >> int_p;

            xmax = root_node_d[str_p("xmax")] >> int_p;

            ymin = root_node_d[str_p("ymin")] >> int_p;

            ymax = root_node_d[str_p("ymax")] >> int_p;

            parameters = root_node_d[str_p("parameters")] >>
                            +parameter >>
                         discard_node_d[str_p("endParameters")];

            parameter = discard_node_d[str_p("parameter")] >> root_node_d[parameterName]  >>
                    discard_node_d[str_p("value")] >> discard_node_d[ch_p('=')] >>
                    real_p;

            parameterName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            basis = root_node_d[str_p("basis")] >>
                        +vectorDef >>
                    discard_node_d[str_p("endBasis")];

            vectorDef = discard_node_d[ch_p('(')] >> vectorComponent >>
                            root_node_d[ch_p(',')] >>
                        vectorComponent >> discard_node_d[ch_p(')')];

            vectorComponent = leaf_node_d[ lexeme_d[+(alnum_p | ch_p('-') | ch_p('/'))]];

            vertices = root_node_d[str_p("vertices")] >>
                        +vectorDef >>
                       discard_node_d[str_p("endVertices")];

            neighbours = root_node_d[str_p("neighbours")] >>
                            +vectorDef >>
                         discard_node_d[str_p("endNeighbours")];

        }

        rule<ScannerT> Lattice, latticeName, dimension, scale, units, unitName, xmin, xmax, ymin, ymax, parameters, parameter, parameterName, basis, vectorDef, vectorComponent, vertices, neighbours;

        rule<ScannerT> const&
        start() const { return Lattice; }
    };
};




struct LPPsystem_p: public grammar<LPPsystem_p>
{
    template <typename ScannerT>
    struct definition
    {
        definition(LPPsystem_p const&)
        {

            LPPsystem = discard_node_d[str_p("LPPsystem")] >> root_node_d[LPPsystemName] >>
                            SPsystems >>
                            lattice >>
                            spatialDistribution >>
                        discard_node_d[str_p("endLPPsystem")];
            
            LPPsystemName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            SPsystems = root_node_d[str_p("SPsystems") ]>>
                            +(SPsystem) >>
                        discard_node_d[str_p("endSPsystems")];

            SPsystem = discard_node_d[str_p("SPsystem")] >> root_node_d[SPsystemName] >> discard_node_d[str_p("from")] >> SPsystemFile;

            SPsystemName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            SPsystemFile = leaf_node_d[lexeme_d[+(alnum_p | ch_p('.'))]];

            lattice = discard_node_d[str_p("lattice")] >> root_node_d[latticeName] >> discard_node_d[str_p("from")] >> latticeFile;

            latticeName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            latticeFile = leaf_node_d[lexeme_d[+(alnum_p | ch_p('.'))]];

            spatialDistribution = root_node_d[str_p("spatialDistribution")] >>
                                    +(positions) >>
                                  discard_node_d[str_p("endSpatialDistribution")];

            positions = discard_node_d[str_p("positions")] >> discard_node_d[str_p("for")] >> root_node_d[SPsystemName] >>
                            !(parameters) >>
                            coordinates >>
                        discard_node_d[str_p("endPositions")];

            parameters = root_node_d[str_p("parameters")] >>
                            +(parameter) >>
                         discard_node_d[str_p("endParameters")];

            parameter = discard_node_d[str_p("parameter")] >> root_node_d[parameterName] >>
                        discard_node_d[str_p("=")] >> int_p >> discard_node_d[str_p(":")] >> int_p >> discard_node_d[str_p(":")] >> int_p;

            parameterName = leaf_node_d[ lexeme_d[+(alnum_p)]];

            coordinates = root_node_d[str_p("coordinates")] >>
                            xCoordinate >>
                            !(yCoordinate) >>
                          discard_node_d[str_p("endCoordinates")];

            xCoordinate = root_node_d[str_p("x")] >> discard_node_d[str_p("=")] >> leaf_node_d[ lexeme_d[+(alnum_p | ch_p('+') | ch_p('-') | ch_p('*') | ch_p('/'))]];
            yCoordinate = root_node_d[str_p("y")] >> discard_node_d[str_p("=")] >> leaf_node_d[ lexeme_d[+(alnum_p | ch_p('+') | ch_p('-') | ch_p('*') | ch_p('/'))]];

        }

        rule<ScannerT> LPPsystem, LPPsystemName, SPsystems, SPsystem, SPsystemName, SPsystemFile, lattice, latticeName, latticeFile, spatialDistribution, positions, parameters, parameter, parameterName, coordinates, xCoordinate, yCoordinate;

        rule<ScannerT> const&
        start() const { return LPPsystem; }
    };
};



std::string generateXML(iter_t tree);
void addAlphabet(xmlpp::Element* alphabetNode, iter_t child);
void addCompartmentLabels(xmlpp::Element* compartmentLabelsNode, iter_t compartments);
void addMembraneStructure(xmlpp::Element* compartmentsNode, iter_t compartments);
void addInitialMultisets(xmlpp::Element* initialMultisetsNode, iter_t initialMultisets);
void addRuleSets(xmlpp::Element* ruleSetsNode, iter_t ruleSets);
void addModule(iter_t ruleMod, xmlpp::Element* ruleSetNode);
void addRule(iter_t ruleMod, xmlpp::Element* ruleSetNode);
void addTranslRule(iter_t ruleMod, xmlpp::Element* ruleSetNode);
std::string generateLibraryXML(iter_t tree);
std::string generateLatticeXML(iter_t tree);
std::string generateLPPXML(iter_t tree);
void addSPsystems(iter_t SPsystemsDeclaration, xmlpp::Node* LPPSystemNode);
void addLattice(iter_t latticeDeclaration, xmlpp::Node* LPPSystemNode);
void addSpatialDistribution(iter_t spatialDistributionDef, xmlpp::Node* LPPSystemNode);
void addParameters(iter_t parametersDef, xmlpp::Node* positionNode);
void addCoordinates(iter_t coordinatesDef, xmlpp::Node* positionNode);
std::string removeSpaces(std::string str);


}

#endif	/* _BNF_H */

