/*
 * $Id: CAPSystem.h,v 1.5 2009/12/18 13:55:18 jpt Exp $ 
 * Date (yymmdd):               090127
 * Version:                     0.1
 * Written by:                  Dr. Francisco J. Romero-Campero
 * Additional contributions by: Jonathan Blakes
 * (in no particular order)     Dr. Hongqing Cao
 *                              Dr. Natalio Krasnogor
 *				Dr. Jamie Twycross
 * Funded by:                   EPSRC grant EP/E017215/1
 * Website:                     http://www.infobiotic.org/
 * Contact:                     nxk@cs.nott.ac.uk
 */

#ifndef CAPSYSTEM_H
#define CAPSYSTEM_H

#include <iostream>
#include <libxml++/libxml++.h>
#include <map>
#include <muParser.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector> 
#include <ctime>
#include <fstream>
#include <time.h>
#include <sbml/SBMLTypes.h>

using namespace std;
using namespace xmlpp;

namespace ECSB {

////////////////////////////////
// Lattice class declarations //
////////////////////////////////

class lattice
{

 public:
  //Constructors
  lattice();
  lattice(std::string latticeFile);
  //lattice(xmlpp::Node* latticeDeclaration);

  //Default Destructor
  ~lattice();

  //Getters
  int getXMax();
  int getYMax();

  //Displayer
  void show();

 private:

  std::string name;
  int dimension;
  std::string units;
  int scale;
  int xmin;
  int xmax;
  int ymin;
  int ymax;
  std::map<std::string,double> parameters;
  std::vector< std::vector<std::string> > basis;
  std::vector< std::vector<std::string> > vertices;
  std::vector< std::vector<std::string> > neighbours;

};

//////////////////////////////////
// Parameter class declarations //
//////////////////////////////////

class parameter {

public:
    //Default constructor
    parameter();
    //Default destructor
    ~parameter();
    //Default copy constructor
    parameter(const parameter& param);

    //Constructor form XML specification
    parameter(xmlpp::Node* parameterDeclaration);

    //Getters
    std::string getName();
    int getLowerBound();
    int getUpperBound();
    int getStep();

private:
    std::string name;
    int lowerBound;
    int upperBound;
    int step;

};


//////////////////////////////////////
// CAPCompartment class declaration //
//////////////////////////////////////

class CAPCompartment{

 public:
     //Default constructor
     CAPCompartment();
     //Default destructor
     ~CAPCompartment();
     //Copy constructor
     CAPCompartment(const CAPCompartment& comp);

     CAPCompartment(xmlpp::Node* compartmentDeclaration);


    //Displayer
    void show();

    //Getters
    string getName();
    unsigned long getNumberId();
    string getLabel();
    string getFather();
    std::vector<std::string> getChildren();

    //Setters
    void setName(std::string compartmentName);
    void setLabel(std::string compartmentLabel);
    void setFather(std::string compartmentFather);
    void setChildren(std::vector<std::string> childrenNames);
    void setNumberId(unsigned long id);

 private:
  std::string name;
  unsigned long numberId;
  std::string label;
  std::string father;
  std::vector<std::string> children;

};

///////////////////////////////////////
// InitialMultiset class declaration //
///////////////////////////////////////

class InitialMultiset{

 public:
  InitialMultiset();
  InitialMultiset(const InitialMultiset& initialMultiset);
  InitialMultiset(xmlpp::Node* initialMultisetDeclaration);
  ~InitialMultiset();

  //Getters
  std::string getLabel();
  std::map<std::string,unsigned long> getMultiset();

  //Setters
  void setLabel(std::string labelName);
  void setObjectMultiplicity(std::string objectName, unsigned long objectMultiplicity);

  void show();

 private:
  std::string label;
  std::map<std::string,unsigned long> multiset;


};

///////////////////////////////
// CAPRule class declaration //
///////////////////////////////

class CAPRule
{

 public:

  //Constructors
  CAPRule();
  //Copy constructors
  CAPRule(const CAPRule& capRule);
  CAPRule(xmlpp::Node* ruleSeclaration);
  ~CAPRule();

  //Displayer
  void showRule();

  //Getters
  std::string getName(){ return name; };
  std::string getType(){ return type; };
  double getConstant(){ return constant; };
  std::string getStringConstant() {return stringConstant; };
  std::vector<std::string> getLhsOutside(){ return lhsOutside; };
  std::vector<std::string> getLhsInside(){ return lhsInside; };
  std::vector<std::string> getLhsTranslocation() { return lhsTranslocation; };
  std::vector<std::string> getRhsOutside(){ return rhsOutside; };
  std::vector<std::string> getRhsInside(){ return rhsInside; };
  std::vector<std::string> getRhsTranslocation(){ return rhsTranslocation; };
  std::vector<int> getVect(){ return ruleVector; };
  std::string getLabel(){ return label; };
  std::string getLabelLhs() { return labelLhs; };
  std::list<std::string> getObjects();

  //Setters
  void setName(std::string str);
  void setType(std::string str);
  void setConstant(double db);
  void setStringConstant(std::string str);
  void setLhsOutside(std::vector<std::string> objects);
  void setLhsInside(std::vector<std::string> objects);
  void setLhsTranslocation(std::vector<std::string> objects);
  void setRhsOutside(std::vector<std::string> objects);
  void setRhsInside(std::vector<std::string> objects);
  void setRhsTranslocation(std::vector<std::string> objects);
  void setRuleVector(std::vector<int> coordinates);
  void setLabelLhs(std::string str);
  void setLabel(std::string str);



  void substituteObjects(std::map<std::string,std::string> substitution);
  void substituteConstants(std::map<std::string,std::string> substitution);
  void substituteConstants(std::map<std::string,double> substitution);
  void substituteLabels(std::map<std::string,std::string> substitution);

 private:
  std::string name;
  std::string type;
  double constant;
  std::string stringConstant;
  std::vector<std::string> lhsOutside;
  std::vector<std::string> lhsInside;
  std::vector<std::string> lhsTranslocation;
  std::vector<std::string> rhsOutside;
  std::vector<std::string> rhsInside;
  std::vector<std::string> rhsTranslocation;
  std::vector<int> ruleVector;
  std::string labelLhs;
  std::string label;

};

//////////////////////////////////
// CAPRuleSet class declaration //
//////////////////////////////////

class CAPRuleSet
{

public:
  //Default constructor
  CAPRuleSet();
  //Copy constructor
  CAPRuleSet(const CAPRuleSet& capRuleSet);
  //Constructor from xml specification
  CAPRuleSet(xmlpp::Node* ruleSetDeclaration);
  //Default destructor
  ~CAPRuleSet();

  //Displayer
  void showRuleSet();

  //Getters
  std::string getLabel(){ return label; };
  unsigned int getNumberRules(){ return numberRules; };
  std::vector<CAPRule> getRules(){ return setOfRules; };
  std::list<std::string> getObjects();

  //Setters
  void setLabel(std::string compartmentName);
  void setRules(std::vector<CAPRule> rules);
  void addRule(CAPRule rule);

 private:
  std::string label;
  unsigned int numberRules;
  std::vector<CAPRule> setOfRules;

};

////////////////////////////////////////
// ModuleDefinition class declaration //
////////////////////////////////////////

class ModuleDefinition
{

  private:

    std::string name;
    std::vector<std::string> objects;
    std::vector<std::string> constants;
    std::map<std::string,double> upperBounds;
    std::map<std::string,double> lowerBounds;
    std::map<std::string,double> precisions;
    std::map<std::string,std::string> scales;
    std::vector<std::string> labels;
    std::vector<CAPRule> rules;

  public:

    //Constructors
    ModuleDefinition();
    ModuleDefinition(const ModuleDefinition& moduleDefinition);
    ModuleDefinition(xmlpp::Node* moduleDeclaration, std::map<std::string,ModuleDefinition> partialLibrary);
    ~ModuleDefinition();

    //Displayer
    void showModule();

    //Getters
    std::string getName();
    std::vector<std::string> getObjects();
    std::vector<std::string> getConstants();
    double getLowerBound (std::string);
    double getUpperBound (std::string);
    double getPrecision (std::string);
    std::string getScale (std::string);
    std::vector<std::string> getLabels();
    std::vector<CAPRule> getRules();

    //Setters
    void setObjects(std::vector<std::string> newObjects);
    void setRules(std::vector<CAPRule> newRules);

    //Instantiating a module definition
    std::vector<CAPRule> instantiate(std::map<std::string,std::string> objectsSubstitution,
				  std::map<std::string,string> constantsSubstitution,
				  std::map<std::string,std::string> labelsSubstitution);


};

/////////////////////////////////////
// ModuleLibrary class declaration //
/////////////////////////////////////

class ModuleLibrary
{

 private:

  std::string name;
  std::map<std::string,ModuleDefinition> modules;

 public:

  //Constructors
  ModuleLibrary();
  ModuleLibrary(const ModuleLibrary& moduleLibrary);
  ModuleLibrary(std::string libraryFile);

  ~ModuleLibrary();

  //Displayer
  void showModuleLibrary();

  //Getter
  std::map<std::string,ModuleDefinition> getModules() { return modules; };


};

///////////////////////////////
// PSystem class declaration //
///////////////////////////////

class PSystem
{
 public:
  //Constructors
  PSystem();
  //Copy constructor
  PSystem(const PSystem& pSystem);
  PSystem(SBMLDocument* sbmlDocument);
  PSystem(xmlpp::Node* pSystemDeclaration);

  //Default Destructor
  ~PSystem();

  //Getters
  std::string getName();
  std::vector<std::string> getAlphabet();
  std::map<std::string,CAPRuleSet> getRuleSets();
  std::map<std::string,InitialMultiset> getInitialMultisets();
  InitialMultiset getInitialMultiset(std::string label);
  unsigned long getNumberOfCompartments();
  std::map<std::string,CAPCompartment> getMembraneStructure();
  std::list<std::string> getRuleObjects();

  //Setters
  void setName(std::string pSystemName);
  void setObjectMultiplicity(std::string label, std::string objectName, unsigned long objectMultiplicity);
  //Displayer
  void show();

  //Converter to translocation rules
  void convertTranslocation();



 protected:
  void createAlphabet(xmlpp::Node::NodeList alphabetDeclaration);
  void createSetOfLabels(xmlpp::Node::NodeList labelsDeclaration);
  void createMembraneStructure(xmlpp::Node::NodeList compartmentsDeclaration);
  std::vector<std::string> findChildren(std::string compartmentName);
  void assignChildren();
  void createInitialMultisets(xmlpp::Node::NodeList initialMultisetsDeclaration);
  void createRuleSets(xmlpp::Node::NodeList ruleSetsDeclaration);
  void checkAlphabet();
  CAPRule convertTranslocationRule(CAPRule rule, std::string compartmentName);

  void createAlphabet(Model* sbmlModel);
  bool isDegradation(std::string s);
  void createMembraneStructure(Model* sbmlModel);
  void createSetOfLabels();
  void createInitialMultisets(Model* sbmlModel);
  void createRuleSets(Model* sbmlModel);
  void checkValid(Model* sbmlModel, Reaction* reaction);
  std::string reactionType(Model* sbmlModel, Reaction* reaction);
  std::vector<std::string> reactants(Model* sbmlModel, Reaction* reaction);
  std::vector<std::string> products(Model* sbmlModel, Reaction* reaction);
  void addRule(std::string compartmentName, CAPRule rule);

 private:
  std::string name;
  std::vector<std::string> alphabet;
  std::vector<std::string> setOfLabels;
  std::map<std::string,CAPCompartment> membraneStructure;
  std::map<std::string,InitialMultiset> initialMultisets;
  std::map<std::string,CAPRuleSet> ruleSets;

};

/////////////////////////////////
// automaton class declaration //
/////////////////////////////////

class automaton
{

 public:
  //Constructors
  automaton();
  automaton(std::vector<int> coordinates, std::string pSystem);
  automaton(std::vector<int> coordinates, std::string pSystem, InitialMultiset multiset);

  //Copy constructor
  automaton(const automaton& aut);

  //Default Destructor
  ~automaton();

  //Getters
  std::vector<int> getPosition();
  std::string getPSystemName();
  InitialMultiset getMultiset();



  //Setters
  void setPosition(std::vector<int> pos);
  void setPSystemName(std::string pSystemName);

  //Displayer
  void show();

 private:

  //int dimension;
  //int[dimension] position;
  std::vector<int> position;
  std::string pSystem;
  InitialMultiset multiset;

};

////////////////////////////////////////
// CellularAutomata class declaration //
////////////////////////////////////////

class CellularAutomata
{

 public:
  //Constructors
  CellularAutomata();
  //Copy constructor
  CellularAutomata(const CellularAutomata& cellularAutomata);
  CellularAutomata(xmlpp::Node* cellAutomataDeclaration, std::map<std::string,PSystem> pSystems);

  //Default Destructor
  ~CellularAutomata();

  //Mutation
  void mutate(double probability, int xLowerValue, int xUpperValue, int yLowerValue, int yUpperValue);

  //Getters
  std::vector<automaton> getAutomata();
  unsigned long getNumberOfCompartments();
  double getNoise();
  int getNumberMutations();


  //Displayer
  void show();

 protected:
  std::vector< std::vector<int> > combineCoordinates(std::vector< std::vector<int> >);
  std::vector< std::vector<int> > combineParametricCoordinates(std::vector< std::vector<int> > coordinatesPrecursor);
  std::vector< std::vector<double> > combineValues(std::vector< std::vector<double> > separateValues);
  std::vector< std::vector <std::vector<int> > > combineRandomCoordinates(std::vector< std::vector<int> >, int);
  void fixedPositionAutomata(xmlpp::Node::NodeList automataDeclaration, std::map<std::string,PSystem> pSystems);
  void randomPositionAutomata(xmlpp::Node::NodeList automataDeclaration, std::map<std::string,PSystem> pSystems);
  std::map< std::vector<int>, ECSB::automaton > convertToMap();
  std::vector<int> neighbouringPosition(std::vector<int> pos);
  std::vector<int> evaluateExpression(std::string expression, std::vector<ECSB::parameter> parameters);

 private:

  std::string name;
  std::vector<automaton> automata;
  unsigned long numberOfCompartments;
  std::string distribution;
  double noise;
  int numberMutations;
  //int dimension;
  //std::map<std::vector<int>,automaton>;

};

/////////////////////////////////
// CAPsystem class declaration //
/////////////////////////////////

class CAPSystem
{

 public:

  //Constructors
  CAPSystem();
  //Copy constructor
  CAPSystem(const CAPSystem& capSystem);
  CAPSystem(std::string fileName);

  //Destructor
  ~CAPSystem();


  //Displayer
  void show();

  //Getters
  std::string getName();
  std::vector<std::string> getCombinedAlphabets();
  std::map<std::string,CAPRuleSet> getCombinedRuleSets();
  std::map<std::string,InitialMultiset> getCombinedInitialMultisets();
  unsigned long getNumberOfCompartments();
  CellularAutomata getCellularAutomata();
  std::map<std::string,PSystem> getPSystems();
  PSystem getPSystem(std::string);

  //Setters
  void setPSystems(std::map<std::string,PSystem> systems);

 protected:
  //void createPSystems(xmlpp::Node::NodeList pSystemsDeclaration);
  std::map<std::string,PSystem> createPSystems(xmlpp::Node::NodeList pSystemsDeclaration);
  void createLattice(xmlpp::Node* latticeDeclaration);
  //void createCellularAutomata(xmlpp::Node* cellularAutomataDeclaration, std::list<std::string> pSystemNames);

 private:

  std::string name;
  std::map<std::string,PSystem> pSystems;
  lattice spatialLattice;
  CellularAutomata cellAutomata;

};

////////////////////////////////////////
// miscellaneos function declarations //
////////////////////////////////////////

//Expand module definitions
std::vector<CAPRule> expand(xmlpp::Node *module);
std::vector<CAPRule> expand(xmlpp::Node *module, std::map<std::string,ModuleDefinition> partialLibrary);

//Remove duplicate rules
std::vector<CAPRule> removeDuplicates(std::vector<CAPRule>);

//Put individual names to rules
std::vector<CAPRule> individualRuleNames(std::vector<CAPRule> rules, std::string lab);

//Checks if a rule is is a vector of rules
int ruleIn(CAPRule rule, std::vector<CAPRule> rules, int pos);

int ruleEqual(CAPRule r1, CAPRule r2);

ModuleDefinition renameVariables(ModuleDefinition module);

}

#endif
