// miscellaneous functions
// copyright 2008, 2009 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#ifndef MISC_H
#define MISC_H

#include <sys/time.h>
#include <math.h>
#include <string>
#include <cassert>
#include <sstream>
#include <iostream>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <errno.h>
#include <muParser.h>
#include <vector>

namespace ECSB {

// generate human readable timestamp string
const char *generateTimestamp(void);

// convert time in seconds to human readable string
const char *convert_time(const unsigned long time);

// convert time in seconds and fractions of seconds to human readable string
const char *convert_time(const double time);

// compute difference in seconds between two timevals
double timeval_diff(struct timeval &from, struct timeval &to);

// compute difference in seconds between two timestamps
double timestamp_diff(const char *from, const char *to);

// return date from timestamp
const char *timestamp_date(char *time);

// return time from timestamp
const char *timestamp_time(char *time);

// convert string to specified type
template <typename T> inline T fromStringTo(const std::string &str)
{
	T result;
	std::istringstream stream(str);
	assert(stream >> result);
	return result;
}

// convert specified type to string
template <typename T> inline std::string somethingToString(const T &t)
{
	std::ostringstream outputStream;
	outputStream << t;
	return outputStream.str();
}

// create subsitution to instantiate modules
std::map<std::string,std::string> createSubstitution(std::vector<std::string> v1, \
		std::vector<std::string> v2);

// convert string with specific delimiters into a vector
std::vector<std::string> convertToVector(std::string str, \
		std::string delimiters = ",");

int equalityVector(std::vector<std::string> v1, std::vector<std::string> v2);

bool isNumeric(std::string);

}

// stochastic comparison event (greater than threshold)
double stochastic_comparison_GT(double E, double X, double T1, double T2);

// stochastic comparison event (greater than or equal to threshold)
double stochastic_comparison_GE(double E, double X, double T1, double T2);

// stochastic comparison event (less than threshold)
double stochastic_comparison_LT(double E, double X, double T1, double T2);

// stochastic comparison event (less than or equal to threshold)
double stochastic_comparison_LE(double E, double X, double T1, double T2);

// initialise stochastic comparison events
void initialise_stochastic_comparisons(mu::Parser &parser);

// reassign stochastic comparison event variables
void update_stochastic_comparisons_variables(mu::Parser &parser);

// free stochastic comparison event variables
void free_stochastic_comparisons_variables(mu::Parser &parser);

#endif
