#!/bin/sh

PKGBASE=$(cat NAME)
VERSION=$(cat VERSION)
PKGNAME=$PKGBASE-$VERSION
CTRLFILE=control
LICENSE="/usr/share/doc/$PKGNAME/COPYING"
README="/usr/share/doc/$PKGNAME/README"
MAINTAINER=$(cat $CTRLFILE | egrep '^%vendor' | sed 's/%vendor //')

echo "*** creating debian package ***"
CWD=$(pwd)
TMPDIR=/tmp/$PKGNAME
TMPFILE="$TMPDIR/$PKGNAME.tmp"
ARCH=$(dpkg --print-architecture)
rm -rf $TMPDIR

echo "*** building source ***"
rm -rf $TMPDIR &&
mkdir $TMPDIR &&
./autogen.sh &&
./configure --prefix=$TMPDIR/build --enable-python=yes &&
make all install &&
echo "*** source built ok ***" || {
	echo "*** error building source ***"
	exit 1
}
# fix paths
STRING=$(echo "$TMPDIR/build" | sed 's/\//\\\//g')
for i in lib/pkgconfig/$PKGBASE.pc lib/$PKGBASE.la
do
	cat $TMPDIR/build/$i | sed "s/$STRING/\/usr/g" > $TMPFILE
	mv $TMPFILE $TMPDIR/build/$i
done

LASTVERSION=$(echo $VERSION | cut -f 3 -d '.')
LASTVERSION=$(echo "${LASTVERSION}-1" | bc)
LASTVERSION="$(echo $VERSION | cut -f1,2 -d '.').${LASTVERSION}"
# create control file
echo "*** creating deb ***"
# create control file header
cp $CTRLFILE $TMPDIR/$PKGNAME.list
cat << EOF >> $TMPDIR/$PKGNAME.list
%product $PKGNAME
%copyright $LICENSE
%version $VERSION
%readme $README
%replaces ${PKGBASE}-${LASTVERSION}
EOF

# create filelist
mkepmlist -g root -u root --prefix /usr $TMPDIR/build >> $TMPDIR/$PKGNAME.list &&
cd $TMPDIR &&
epm -a $ARCH --output-dir . -v -f deb -n $PKGNAME &&
mv ${PKGNAME}-$VERSION.deb $CWD/${PKGNAME}_${VERSION}_$ARCH.deb &&
cd $CWD &&
echo "*** binary package build ok ***" || {
	echo "*** error creating binary package ***"
	exit 1
}

# make source package
make dist &&
cd $TMPDIR &&
#cp $PKGBASE-$VERSION.tar.gz ./ &&
tar -xzvf $CWD/$PKGBASE-$VERSION.tar.gz &> /dev/null &&
mv $PKGBASE-$VERSION $PKGBASE-$VERSION-$VERSION
cd $PKGBASE-$VERSION-$VERSION &&
dpkg -e $CWD/${PKGNAME}_${VERSION}_$ARCH.deb &&
mv DEBIAN debian &&
cat debian/control | egrep '(^Package)|(^Maintainer)|(^Description)' > debian/1 &&
mv debian/1 debian/control &&
cat << EOF >> debian/control
Source: $PKGNAME
Architecture: any
Section: science
Priority: optional
EOF
cat << EOF > debian/changelog
$PKGNAME ($VERSION) unstable; urgency=low
  * comment here.

 -- $MAINTAINER  $(date '+%a, %d %b %Y %T %z')
EOF
cd $TMPDIR &&
dpkg-source -b $PKGBASE-$VERSION-$VERSION &&
mv $PKGBASE-${VERSION}_$VERSION.tar.gz $PKGBASE-${VERSION}_$VERSION.dsc $CWD/ &&
cd $CWD &&
echo "*** source package build ok ***" || {
	echo "*** error creating source package ***"
	exit 1
}

# make meta package
# create control file
mkdir -p $TMPDIR/$PKGBASE/DEBIAN
cat << EOF > $TMPDIR/$PKGBASE/DEBIAN/control
Package: $PKGBASE
Version: $VERSION
Section: science
Priority: optional
Architecture: all
Depends: $PKGNAME
Replaces: $PKGBASE (<< $VERSION)
EOF
cat $CTRLFILE | egrep '^%vendor' | sed 's/%vendor/Maintainer:/' >> $TMPDIR/$PKGBASE/DEBIAN/control
cat $CTRLFILE | egrep '^%description' | sed 's/%description/Description:/' >> $TMPDIR/$PKGBASE/DEBIAN/control

# make meta package
cd $TMPDIR &&
dpkg-deb --build $PKGBASE &&
mv $PKGBASE.deb $CWD/${PKGBASE}_${VERSION}_all.deb &&
cd $CWD &&
echo "*** meta package build ok ***" || {
	echo "*** error creating meta package ***"
	exit 1
}

# remove tmp files
rm -rf $TMPDIR

echo "*** all packages built ok ***"
